#!/usr/bin/perl
#!/usr/local/gnu/bin/perl
use strict;

# Author: Dong Wang (dongw@cs.cmu.edu)

# convert a structural netlist from ver to smv model. We also generate a
# top main module with the ivf module instantiated.
# $pins is the nuber of pots of top module, $pinarray, $modearray are the 
# port and its direction. 

# will generate a smv file and also a mapping file, from names in smv to
# original names from ivf

# given the pred signals from design compiler, which is
# muxselect={"...",...} ifcond={"...",...}, we say in the i2s file, what 
# are the smv signals that are the predicates

#$main = 1;

if ($#ARGV != 2 && $#ARGV != 3) {
  print "usage: ivf2smv.pl ivffile case_if_pred [goalfile] smvfile\n";
  print "generate a smv file and a mapping file from smv signal name to ivf signal name and optionally the signals in the property, generate a smv and a mapping from verilog signal to smv name file which also include the extracted predicates. If a goalfile is provided, also generate smvfile.goal to hold the mapped to smv signals\n";
  exit(0);
}

my ($ivffile, $predfile, $goalfile, $smvfile);
my $withcoi = 1;
if ($#ARGV == 3) {
  ($ivffile, $predfile, $goalfile, $smvfile) = @ARGV;
} else {
  $goalfile = "";
  $ivffile = $ARGV[0];
  $predfile = $ARGV[1];
  $smvfile = $ARGV[2];
  $withcoi = 0;
}

# Find the design name in the file
if(!(-r $ivffile)) {
  printf("Can't open file %s\n",$ivffile);
  exit(-1);
}
open(IVFFILE,$ivffile);
if(!(-r $predfile)) {
  printf("Can't open file %s\n",$predfile);
  exit(-1);
}
open(PREDFILE,$predfile);

my $sigmapfile;

($sigmapfile = $smvfile) =~ s/\.smv/.i2s/;

my %logicname = ("AND", 0, "OR", 1, "XOR", 2, "NOT", 3, "LATCH_SR", 4, "NL_GATE", 5, "MUX", 6);
my @smvoperator = ("&", "|", "xor", "!", "", "=", "");

open(SMVFILE,">".$smvfile);
open(SIGMAPFILE,">".$sigmapfile);


# the index of how many mapped input sig has been processed
my $in_num = 0;
my $out_num = 0;
# the index of how many reg/defined sig has been processed
my $g_num = 0;
my $reg_num = 0;

my $clk = "";
my  $set = "";
my  $clear = "";

my %gates = ();
my @pinarray = ();
my @modearray = ();
my $designname;
# Set the line separator to a semicolon
$/ = '}';
my %fsa_map = (); # the alias should be substituted, esp when using cyco -w
#acutally only support one module now
while (&subst_a_module()) {
}
$/ = "\n";

my %goals = ();
my %cared = ();
if ($withcoi) {
  &read_property_sig($goalfile);
  %goals = %cared;
  &coi_analysis();
}

my %sigmap = ();
&verilog2smvname();

my $previous = "";
&dump_smv();

#read the pred file and build @preds
my %preds = ();
&read_pred();

my $num_pred = 0;
my @smvpred = (); 
foreach (sort keys %preds) {
  my $pred = "$_/0";
  if ((exists $sigmap{$pred})) {
    $smvpred[$num_pred] = $sigmap{$pred};
    $num_pred = $num_pred + 1;
  }
}
printf ("There are %d predicates after design compiler, %d predicates after tosmv\n", scalar(keys %preds), $num_pred);
printf SIGMAPFILE ("%d\n", $num_pred);
my $i;
for ($i = 0; $i < $num_pred; $i++) {
  printf SIGMAPFILE ("%s\n", $smvpred[$i]);
}

# dump the sigmap to the file
printf SIGMAPFILE ("%d\n", scalar(keys %sigmap));
foreach (sort keys %sigmap) {
  my $ivfname = $_;
  my $smvname = $sigmap{$_};
  printf SIGMAPFILE ("%s <-> %s\n", $smvname, $ivfname);
}
if ($ivffile != "") {close(IVFFILE);}
close(SMVFILE);
close(SIGMAPFILE);

if ($withcoi) {
  my $newgoal;
  ($newgoal = $smvfile) =~ s/\.smv/.goal/;
  open(NEWGOAL,">".$newgoal);
  foreach (keys %goals) {
    printf NEWGOAL ("%s.0\n", $sigmap{$_});
  }
  close(NEWGOAL);
}

#print out some statiscis
print "smv model has $reg_num registers, $g_num gates\n";



#my $rec = {};
# OUTPUT => $string,
#	TYPE => $string,
#	INPUTS  => [ @old_values ],
#	FLAG => $number,
 
sub subst_a_module {
  #($sigmap) = @_;

  my $theline = <IVFFILE>; $theline =~ s/^\s+|\s+$|\n//g;
  if (!$theline) {
    return 0;
  }

  # find the MX
  while (!($theline =~ /^MX/)) {
    $theline = <VLOGFILE>;
  if (!$theline) {
    #end of file
    return 0;
  }
    $theline =~ s/^\s+|\s+$|\n//g;
  }
  
  # Grab everything on the line before the first "("
  my $firsthalf = $theline;
  $firsthalf =~ s/\n//g;
  $firsthalf =~ s/^([^\{]+).*/\1/;
  #printf "firsthalf $firsthalf\n"; #debug
  # Remove whitespace from beginning of line
  $firsthalf =~ s/^\s+//;

  my @firsts = split(/\s+/,$firsthalf);
  if (!($firsts[$#firsts-1] eq "MX")) {
    print @firsts;
    print " is bad header of a module\n";
    exit(1);
  }
  $designname = $firsts[$#firsts];
  $designname =~ s/^(.*)(\.logic)/\1/;
  if ($designname eq "") {
    print "$firsts[$#firsts] is bad name of a module\n";
    exit(1);
  }
  printf(" module %s\n",$designname);

  # Grab everything on the line after the "{MP{"
  my $lasthalf = $theline;
  $lasthalf =~ s/\n//g;
  $lasthalf =~ s/([^\{]+)(\{MP\{)(.*);\}/\3/;
  #print "lasthalf $lasthalf\n", #debug
  my $pinlist = $lasthalf;
  $pinlist =~ s/^\s+|\s+$//g;
  # The pinlist should now be a	comma-separated list.
  my @pinmodearray = split(/;/,$pinlist);
  my $pins = 0;
  

  for($i=0;$i<=$#pinmodearray;$i++) {
    my $newpin;
    $pins = $pins + 1;
    #print "pinmodearray[i] $pinmodearray[$i]\n"; #debug
    my ($pin, $mode) = split(/\s+/,$pinmodearray[$i]);
    $pin = &regular_name($pin);
    if ($mode ne "OUT" && $mode ne "IN") {
      printf("invalid mode %s\n", $mode);
      exit(1);
    }
    push @pinarray, $pin;
    push @modearray, $mode;
  }

  printf("toplevel ports %d\n", $#pinmodearray);

  # the } after MX
  $theline = <IVFFILE>; $theline =~ s/^\s+|\s+$|\n//g;
  # the first line of signals
  $theline = <IVFFILE>; $theline =~ s/^\s+|\s+$|\n//g;
  $theline =~ s/^MF\{//;
  # skip the signals

  while($theline =~  /^FSA/ || $theline =~ /^FS/) {

#    if ($theline =~ /^FSA/) {
#      if ($theline =~ /^FSA\{(\S+)\s*(\S+)\s*\}$/) {
#	$fsa_map{$1} = $2;
#      } else {
#	print "bad fsa $theline\n";
#      }
#    }
    $theline = <IVFFILE>; $theline =~ s/^\s+|\s+$|\n//g;
  }


  while ($theline =~ /^(FG)/) {
    if ($theline =~ /^FG ([A-Z_]+)\.(\d+){[A-Z_]+ 0 I (.+)}$/) {
      # $1 is the NOT/AND/OR ..
      # $2 is the gate number
      # $3 is the pins
      my $gtype = $1;
      my $gnum = $2;
      my $gpins = $3;
      #print "gtype $gtype gnum $gnum gpins $gpins\n"; #debug
      if (!(exists $logicname{$gtype})) {
	printf("gate type is not supported $theline\n");
	exit(1);
      }
      my $index = $logicname{$gtype};
      my @gpinarray = split(/\s+/,$gpins);
      # $0 become 0, $1 become 1
      for ($i = 0; $i <= $#gpinarray; $i++) {
	if ($gpinarray[$i] eq "\$0") {
	  $gpinarray[$i] = "0";
	} elsif ($gpinarray[$i] eq "\$1") {
	  $gpinarray[$i] = "1";
	} else {
	  $gpinarray[$i] = &regular_name($gpinarray[$i]);

	}
      }
      #print "gpinarray @gpinarray\n"; #debug
      my $num_in = $gpinarray[0];
      my $num_out = $gpinarray[$num_in+1];
      my $output = $gpinarray[$num_in+2];
      #$newoutput = &replace_name($output, 1);
      
      if ($num_out != 1) {
	print "the number of output of the gate is not 1, $theline\n";
	exit(1);
      }

      my $onegate = {};
      $onegate->{OUTPUT} = $output;
      $onegate->{TYPE} = $gtype;
      $onegate->{FLAG} = 0;
      $onegate->{INPUTS} = [ @gpinarray[1..$num_in] ];

      $gates{$onegate->{OUTPUT}} = $onegate;
    } else {
      print "bad gate declartion $theline\n";
      exit(1);
    }
    $theline = <IVFFILE>; $theline =~ s/^\s+|\s+$|\n//g;
    #print "theline $theline\n"; #debug
  }
  if ($theline ne "\}") {
    print "bad end of module $theline\n";
    return (0);
  }

  return 1;
}

sub dump_smv {
  printf SMVFILE ("-- %s\nMODULE main\n", $designname);
  printf SMVFILE ("--input output vars\nVAR\n");
  # print out the in ports
  for ($i = 0; $i <= $#pinarray; $i++) {
    my $pin = $pinarray[$i];
    my $mode = $modearray[$i];
    my $newpin = &newname($pin);
    if ($mode eq "IN") {
      printf SMVFILE ("%s: boolean; -- %s\n", $newpin, $pin);
    } else {
      printf SMVFILE ("--%s: boolean; -- %s\n", $newpin, $pin);
    }
  }
  foreach (keys %gates) {
    my $onegate = $gates{$_};
    if (!$withcoi || $onegate->{FLAG} == 1) {
      &print_gate($onegate);
    }
  }

  return 1;
}

# print the gate as what it is from the input file
sub echo_gate {
  my ($onegate) = @_;
  printf ("%s I %d", $onegate->{TYPE}, scalar @{ $onegate->{INPUT}});
  print @ {$onegate->{INPUT}};
  printf ("O 1 %s\n", $onegate->{OUTPUT});
}

sub newname {
  my ($name) = @_;
  if (&trivial_net($name)) {
    return $name;
  }

  my $newname = $sigmap{$name};
  if (defined $newname) {
    return $newname;
  } else {
    print "$name is not mapped\n";
    exit(1);
  }
}

sub print_gate {
  my ($onegate) = @_;

  my $newoutput = &newname($onegate->{OUTPUT});
  my $gtype = $onegate->{TYPE};

  if ($gtype eq "LATCH_SR") {
    if ($clk eq "") {
      $clk = $onegate->{INPUTS}[3];
    }
    if ($onegate->{INPUTS}[3] ne $clk) {
      printf("error: clk %s is not the same as in \n", $clk, &echo_gate($onegate));
      exit(1);
    }
    printf SMVFILE ("VAR %s: boolean; -- %s\n", $newoutput, $onegate->{OUTPUT});
    printf SMVFILE ("ASSIGN\n");
    my $myset = $onegate->{INPUTS}[1];
    my $myreset = $onegate->{INPUTS}[2];
    if ($myset ne "0" || $myreset ne "0") {
      if ($myset ne "0" && $myreset ne "0") {
	printf("warning: reg %s output 's set %s clear %s are both connected\n", $myset, $myreset);
	exit(1);
      } elsif ($myset ne "0") {
	if ($set eq "") {
	  $set = $myset;
	}
	if ($set ne $myset) {
	  printf("warning: reg 's set %s is different with %s\n", $myset, $set);
	  #exit(1);
	}
	printf SMVFILE ("init(%s) := 1;\n", $newoutput);
      } elsif ($myreset ne "0") {
	if ($clear eq "") {
	  $clear = $myreset;
	}
	if ($clear ne $myreset) {
	  printf("warning: reg 's clear %s is different with %s\n", $myreset, $clear);
	  #exit(1);
	}
	printf SMVFILE ("init(%s) := 0;\n", $newoutput);
      }
    } else {
      printf SMVFILE ("init(%s) := 0;\n", $newoutput);
    }
    printf SMVFILE ("next(%s) := %s;\n", $newoutput, &newname($onegate->{INPUTS}[0]));
  } else {
    # this is going to a defined gate
    if ($previous eq "" || $previous eq "LATCH_SR") {
      printf SMVFILE ("DEFINE\n");
    }
    if ($gtype eq "MUX") {
      my $S = &newname($onegate->{INPUTS}[0]);
      my $T = &newname($onegate->{INPUTS}[1]);
      my $F = &newname($onegate->{INPUTS}[2]);
      printf SMVFILE ("  %s := (%s & %s) | (!%s & %s)", $newoutput,
		      $S, $T, $S, $F);
    } elsif ($gtype eq "XOR" && ((scalar @ { $onegate->{INPUTS} }) == 2)) {
      my $in1 = &newname($onegate->{INPUTS}[0]);
      my $in2 = &newname($onegate->{INPUTS}[1]);
      printf SMVFILE ("  %s := (!%s & %s) | (%s & !%s)", $newoutput,
		      $in1, $in2, $in1, $in2);
    } else {
      printf SMVFILE ("  %s := ", $newoutput);
      my $index = $logicname{$gtype};
      my $op = $smvoperator[$index];
      my $num_index = scalar @ { $onegate->{INPUTS} };
      if ($num_index == 1) {
	if ($op eq "!") {
	  printf SMVFILE ("%s %s", $op, &newname($onegate->{INPUTS}[0]));
	} elsif ($op eq "=") {
	  printf SMVFILE ("%s", &newname($onegate->{INPUTS}[0]));
	} else {
	  printf("bad number of pins ",);
	  echo_gate($onegate);
	  exit(1);
	}
      } else {
	for ($i = 0; $i < $num_index; $i++) {
	  printf SMVFILE ("%s", &newname($onegate->{INPUTS}[$i]));
	  if ($i+1 != $num_index) {
	    printf SMVFILE (" %s ", $op);
	  }
	}
      }
    }
    printf SMVFILE (";\n");
  }
  $previous = $gtype;
}


#mode 0 for input, 1 for defined, 2 for output, 3 reg
sub replace_name {
  my ($old, $type) = @_;
  if ($old eq "0" || $old eq "1") {
    return $old;
  }
  my $newpin;
  if (!(exists $sigmap{$old})) {
    if ($type == 0) {
      $in_num++;
      $newpin = "in$in_num";
    } elsif ($type == 1) {
      $g_num++;
      $newpin = "g$g_num";
    } elsif ($type == 2) {
      $out_num++;
      $newpin = "out$out_num";
    } elsif ($type == 3) {
      $reg_num++;
      $newpin = "reg$reg_num";
    } else {
      print "error: bad type $type in calling to replace_name\n";
      exit(1);
    }
    $sigmap{$old} = $newpin;
  } else {
    $newpin = $sigmap{$old};
  }
  return $newpin;
}
# make the name to be a path/bit
sub regular_name {
  my ($old) = @_;
  my $unalias = $fsa_map{$old};
  if (defined $unalias) {
    $old = $unalias;
  }
  $old =~ s/\\//g;# remove the \ added by ver
  if ($old =~ /^\w*\/(.*)$/) {#change TOP/s to s, assume TOP is alpha numerical
    $old = $1;
  }
  if ($old =~ /^(.*)(<)(\d+)(>\/0)$/) { #change <5>/0 to /5
    $old = "$1/$3";
  }
  return $old;
}

sub read_pred {
  my $j;
  for ($j = 0; $j < 2; $j++) {
    my $theline = <PREDFILE>; $theline =~ s/^\s+|\s+$|\n//g;
    if (!$theline) {
      printf "incomplete pred file\n";
      exit(1);
    }
    if ($theline =~ /^[a-z]+\s*=\s*{(.*)}$/) {
      my @predarray = split(/,\s+/,$1);
      for ($i = 0; $i <= $#predarray; $i++) {
	$predarray[$i] =~ s/\"//g;
	if (!($predarray[$i] =~ /.*\*Logic0\*|.*\*Logic1\*/)) {
	  if (!(exists $preds{$predarray[$i]})) {
	    $preds{$predarray[$i]} = 1; 
	  }
	}
      }
    } else {
      printf "bad pred $theline\n";
      exit(1);
    }
  }
}

sub trivial_net {
  my ($name) = @_;
  if ($name eq "" || $name eq "0" || $name eq "1") {
    return 1;
  } else {
    return 0;
}
}

# given an array of cared signals, and the gates, mark all gates in the coi
sub coi_analysis {
  my $changed = 1;
  my %newsig = %cared;
  my $onegate = {};
  
  #print %gates; print "\n";

  while ($changed) {
    $changed = 0;
    my %newnewsig = ();

    #print %newsig; print "\n";
    foreach (keys %newsig) {
      $onegate = $gates{$_} ;
      if ((defined $onegate)) {
	$onegate->{FLAG} = 1;
	foreach (@{ $onegate->{INPUTS} }) {
	  my $name = $_;
	  if (&trivial_net($name) == 0 && !(exists $cared{$name})) {
	    $newnewsig{$name} = 1;
	  }
	}
      }
    }
    if ((scalar (keys %newnewsig)) > 0) {
      $changed = 1;
      foreach (keys %newnewsig) {
	$cared{$_} = 1;
      }
    }
    %newsig = %newnewsig;
    %newnewsig = ();
  }
}

sub read_property_sig {
  my ($goalfile) = @_;
  open (GOALFILE, $goalfile);
  while (<GOALFILE>) {
    my $theline = $_; $theline =~ s/^\s+|\s+$//g;
    $cared{$theline} = 1;
  }
  close(GOALFILE);
}

# with coi, a gate 's flag is 1; without it, every gate should be mapped; 
# we also mark all inputs/outputs
sub verilog2smvname {
  # first map the register's output
  my $onegate = {};
  foreach (keys %gates) {
    $onegate = $gates{$_};
    if ($onegate->{TYPE} eq "LATCH_SR" && (!$withcoi || $onegate->{FLAG} == 1)) {
      &replace_name($_, 3);
    }
  }
  # second rest of the gates
  foreach (keys %gates) {
    $onegate = $gates{$_};
    if ($onegate->{TYPE} ne "LATCH_SR" && (!$withcoi || $onegate->{FLAG} == 1)) {
      &replace_name($_, 1);
    }
  }
  # inputs/outputs
  for ($i = 0; $i <= $#pinarray; $i++) {
    my $pin = $pinarray[$i];
    my $mode = $modearray[$i];
    if ($mode eq "IN") {
      &replace_name($pin, 0);
    } else {
      &replace_name($pin, 2);
    }
  }
}
