#!/usr/bin/perl
#!/usr/local/gnu/bin/perl

# Author: Dong Wang (dongw@cs.cmu.edu)

# convert a structural netlist from synthesis to one where the simple strutural
# gates are become verilog logic gates. This is used to save the amount of
# hierarchy. The output of this is given to ver, so I could replace the
# flipflop to ver's latch too.

%structname = ("AN2", 0, "OR2", 1, "XOR2", 2, "INV", 3, "MUX2_1", 4, "TRI_BUF2", 5, "DFF", 6, "SDFF", 7, "D_LATCH", 8);
@num_inputs = (2, 2, 2, 1, 3, 2, 4, 4, 2);
@num_outputs = (1, 1, 1, 1, 1, 1, 2, 1, 1);
@logicname = ("and", "or", "xor", "not", "assign", "bufif1", "latch_sr", "latch_sr", "null");
@ports = ("Z:A:B", "Z:A:B", "Z:A:B", "Z1:A", "Y:S:B:A", "Z:A:E", "Q:D:SD:CD:CP", "Q:D:SD:CD:CP", "");
@supported = (1, 1, 1, 1, 1, 1, 1, 1, 0);

# used to make up a name for a dummy signal, it must be unique (good luck).
$fakename = "fake_dffq";
$fakecounter = 0;

$vlogfilename = $ARGV[0];
$deffilename = $ARGV[1];
# Find the design name in the file
if(!(-r $vlogfilename)) {
  printf("Can't open file %s\n",$vlogfilename);
  exit(-1);
}
open(VLOGFILE,$vlogfilename);
open(DEFFILE,">".$deffilename);
# Set the line separator to a semicolon
$/ = ';';
$firstmodule = 1;

while (&subst_a_module()) {
}

close(VLOGFILE);
close(DEFFILE);


sub subst_a_module {
  if ($firstmodule) {
    $theline = <VLOGFILE>; $otheline = $theline;
    if (!$theline) {
      return 0;
    }
    $theline =~ s/^\s+|\s+$//g;
    $firstmodule = 0;
  }

while (!($theline =~ /module/)) {
  print DEFFILE $otheline;
  $theline = <VLOGFILE>; $otheline = $theline;
  if (!$theline) {
    #end of file
    return 0;
  }
  $theline =~ s/^\s+|\s+$//g;
}

# Grab everything on the line before the first "("
$firsthalf = $theline;
$firsthalf =~ s/\n//g;
$firsthalf =~ s/^([^\(]+).*/\1/;
# Remove whitespace from beginning of line
$firsthalf =~ s/^\s+//;

@firsts = split(/\s+/,$firsthalf);
if (!($firsts[$#firsts-1] eq "module")) {
  print @firsts;
  print " is bad header of a module\n";
  exit(1);
}
$designname = $firsts[$#firsts];

# Grab everything on the line after the first "("
$lasthalf = $theline;
$lasthalf =~ s/([^\(]+)(.*;)/\2/;
#printf("Converting verilog file %s into logic verilog file %s\n",$vlogfilename,$deffilename);
printf(" module %s\n",$designname);

#print the module line out
print DEFFILE $otheline;

# Grab the pins from the module
$done = 0;
$pins = 0;
while(!$done) {
  $theline = <VLOGFILE>; $otheline = $theline;
  $theline =~ s/^\s+|\s+$//g;
  if($theline =~ /^\s*wire/) {
    $done = 1;
  }
  else {
    if(($theline =~ /^\s*input/) ||
       ($theline =~ /^\s*output/) ||
       ($theline =~ /^\s*inout/)) {
      if($theline =~ /\[/) {
	# This is a vector, so grab the size
	$size = $theline;
	$size =~ s/[^\[]+\[//;
	$size =~ s/\].*//;
	($left,$right) = split(/:/,$size);
	# Grab the pin name list
	$pinlist = $theline;
	$pinlist =~ s/[^\]]+\]//;
	$pinlist =~ s/;//;
	$pinlist =~ s/\s//g;
	if($left < $right) {
	  $min = $left;
	  $max = $right;
	}
	else {
	  $min = $right;
	  $max = $left;
	}
	@pinarray = split(/,/,$pinlist);
	for($i=0;$i<=$#pinarray;$i++) {
	  $pins = $pins + ($max - $min + 1);
	  $minarray{$pinarray[$i]} = $min;
	  $maxarray{$pinarray[$i]} = $max;
	}
      }
      else {
	# This is a single bit list, grab the pin	name list
	# Grab the pin name list
	$pinlist = $theline;
	$pinlist =~ s/^\s*input//;
	$pinlist =~ s/^\s*output//;
	$pinlist =~ s/^\s*inout//;
	$pinlist =~ s/;//;
	$pinlist =~ s/^\s+|\s+$//g;
	# The pinlist should now be a	comma-separated list.
	@pinarray = split(/,/,$pinlist);
	for($i=0;$i<=$#pinarray;$i++) {
	  $pins = $pins + 1;
	  $minarray{$pinarray[$i]} = -1;
	  $maxarray{$pinarray[$i]} = -1;
	}
      }
    } else {
      $done = 1;
    }
  }
  if (!$done) {
    print DEFFILE $otheline;
  }
}
# Print out the pins
printf(" ports:%d\n",$pins);
foreach $pinname (keys(%minarray)) {
  if($minarray{$pinname} == -1) {
    #print  " - %s + NET %s ;\n",$pinname,$pinname;
  }
  else {
    for($i=$minarray{$pinname};
	$i<=$maxarray{$pinname}; $i++) {
      #print " - %s[%d] + NET %s[%d] ;\n",$pinname,$i,$pinname,$i;
    }
  }
}

#process the wires
while($thisline =~  /^\s*wire/) {
  print DEFFILE "$otheline\n";
  $theline = <VLOGFILE>; $otheline = $theline;
  $theline =~ s/^\s+|\s+$//g;
}

# Count Components, Build Component Array, Build Net Array
$components = 0;
$flipflop = 0;
while(1) {
  if ($theline =~ /^endmodule/) {
    # the curret module finished
    print DEFFILE "endmodule\n";
    print  "endmodule // $designname $components gates, $flipflop flipflops\n";
    $theline =~ s/endmodule//;
    $otheline = $theline;
    $theline =~ s/^\s+|\s+$//g;
    return (1);
  } elsif($theline =~ /\(/) {
    # Remove any newlines from the line
    $theline =~ s/\n//g;
    # Remove any comments from the line
    $theline =~ s/\/\*[^\*]+\*\///g;
    # Copy the line and remove anything after    the first "("
    $firsthalf = $theline;
    $firsthalf =~ s/([^\(]+).*/\1/;
    #print "firsthalf $firsthalf";
    # Grab the component name and instance name
    ($componentname,$instname) =
      split(/\s+/,$firsthalf);
    #print "instname $instname\n";
    # Bind the instances to components with an associative array
    if (!(exists $structname{$componentname})) {
      print DEFFILE "$otheline\n";
    }
    else {
      my %pinnet;
      $index = $structname{$componentname};
      if ($supported[$index] == 0) {
	print "error: instantiation of $componentname is not support\n";
	exit(1);
      }
      # Copy the line and remove anything    before the first "("
      $lasthalf = $theline;
      $lasthalf =~ s/[^\(]+//;
      # Remove any spaces
      $lasthalf =~ s/\s+//g;
      # Remove first and last parentheses
      $lasthalf =~ s/^\(//;
      $lasthalf =~ s/\);$//;
      # We should then have a comma separated list
      # Split the list into parts
      @pinbinding = split(/,/,$lasthalf);
      #print "lasthalf $lasthalf\n";
      
      $index = $structname{$componentname};
      if ($#pinbinding+1 != $num_inputs[$index]+$num_outputs[$index]) {
	if ($componentname eq "DFF" && 
	    $#pinbinding+2 == $num_inputs[$index]+$num_outputs[$index]) {
	  # we allow QN of DFF not collected, in this case change it to SDFF
	  $componentname = "SDFF";
	  $index = $structname{$componentname};
	} else {
	  print "numb of inputs and outputs are $num_inputs[$index]+$num_outputs[$index], numb of pins are $#pinbinding+1\n";
	  print "warning: some pins are not connected $otheline\n";
	  }
      }
      
      for($i=0;$i<=$#pinbinding;$i++) {
	($pinname,$netname) =
	  split(/\(/,$pinbinding[$i]);
	$pinname =~ s/\.//;
	$netname =~ s/\)//;
	if ($pinname eq "" or $netname eq "") {
	  print "error: bad port map $otheline\n";
	  exit(1);
	}
	$pinnet{$pinname} = $netname;
      }
      @pins = split(/:/,$ports[$index]);
      $newname = $logicname[$index];

      # for SDFF, we require the Q is a signal from the verilog code
      if ($componentname eq "SDFF" or $componentname eq "DFF") {
	$flipflop++;
	if ($instname =~ /^(.+)_reg$/) {
	  $q = $1;
	} elsif ($instname =~ /^(.+)_reg<(\d+)>$/) {
	  $q = "$1<$2>";
	} elsif ($instname =~ /^(.+)_reg<(\d+)\]\[(\d+)>$/) {
	  $q = "$1<$2><$3>";
	} else {
	  $q = "";
	  print "warning: $instname should not be a name of a register from design compiler\n";
	}
	if ($q ne "" && $pinnet{"Q"} ne $q) {
	  $fakeq = $pinnet{"Q"};
	  $pinnet{"Q"} = $q;
	  # build two inverters
	  $fakesig = $fakename."_".$fakecounter;
	  print DEFFILE "not ($fakesig , $q ), ($fakeq , $fakesig );\n";
	  $fakecounter++;
	}
      }


      if ($newname eq "assign") {
	if ($componentname ne "MUX2_1") {
	  print "error: so far only mux is converted to assign\n";
	  exit(1);
	}
	print DEFFILE "assign ";
      } else {
	print DEFFILE "$newname $instname (";
      }

      # look at the require pins names of the logic gate
      for ($i = 0; $i <= $#pins; $i++) {
	if (!(exists $pinnet{$pins[$i]})) {
	  print "error: $pins[$i] in $otheline is not connect\n";
	  exit(1);
	}
	print DEFFILE "$pinnet{$pins[$i]} ";
	# This is converted to an assign statement
	if ($newname eq "assign") {
	  #this is a mux
	  if ($i == 0) {
	    print DEFFILE " = ";
	  } elsif ($i == 1) {
	    print DEFFILE " ? ";
	  } elsif ($i == 2) {
	    print DEFFILE " : ";
	  } else {
	  }
	} else {
	  if ($i == $#pins) {
	    print DEFFILE ")";
	  } else {
	    print DEFFILE ", ";
	  }
	}
      }
      print DEFFILE ";\n";
      # taken care of the other output if any
      if ($num_outputs[$index] == 2) {
	if ($componentname ne "DFF") {
	  print "error: so far on ly DFF has two outputs\n";
	  exit(1);
	}
	if (exists $pinnet{"QN"}) {
	  $qn = $pinnet{"QN"};
	  $q = $pinnet{"Q"};
	  print DEFFILE "assign $qn = ~ $q ;\n";
	}
      }
    }
    $components = $components + 1;
  } else {
    print DEFFILE "$otheline\n";
  }
  if (!($theline = <VLOGFILE>)) {
    print "error: endofmodule is not found\n";
    exit(1);
  }
  $otheline = $theline;
  $theline =~ s/^\s+|\s+$//g;
}

return (1);
}
